// ignore_for_file: public_member_api_docs, sort_constructors_first

import 'dart:async';
import 'dart:developer';
import 'dart:math' as math;

import 'package:flutter/cupertino.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';

typedef FieldMap = Map<String, GlobalKey<FormFieldState>>;
typedef FormMap = Map<String, String>;

class FormxException implements Exception {
  FormxException([String? tag])
      : message = 'FORM.INVALID${tag == null ? '' : '.$tag'}'.toUpperCase();

  final String message;
}

///Controls the [Formx]. Useful for managing field states in another controller.
class FormxController {
  /// All [FormState]. The root [Formx.tag] must be null.
  final states = <String?, Set<FormState>>{};

  /// A [Map] of [FormState]. Where key is the [Formx] tag parameter.
  ///
  /// The main [Formx] tag must be null. Only tag nested [Formx].
  Map<String?, Set<FormState>> get forms {
    assert(
      states.isNotEmpty,
      'Formx not attached, put a [Field] widget below [Formx]',
    );
    return states;
  }

  /// The form as [Json]. (includes nested).
  Map<String, dynamic>? get form {
    if (states.isEmpty) return null;
    final Map<String, dynamic> map = {};

    forms.forEach((key, value) {
      if (key == null) {
        for (var e in value) {
          map.merge(e._scope.form);
        }
      } else {
        for (var e in value) {
          map.merge(e._scope.form.nestIn(key));
        }
      }
    });

    return map;
  }

  /// The [Formx] loading listenable.
  ValueListenable<bool>? get loading => states[null]?.first._scope.isLoading;

  ///Resets all fields and nested [Formx].
  void reset() {
    forms.forEach((map, states) {
      for (var state in states) {
        state.reset();
      }
    });
  }

  ///Validates all fields and nested [Formx], if any.
  bool validate() {
    var isValid = true;
    forms.forEach((map, states) {
      for (var state in states) {
        isValid = state.validate();
      }
    });
    return isValid;
  }

  ///Validates all selects fields with the enclosing [tags].
  bool validateSelected(List<String> tags) {
    final areValid = tags.map(validateTag);
    return !areValid.contains(false);
  }

  ///Validates only the field with the enclosing [tag].
  ///If the tag is a form. Validates all tags below it.
  bool validateTag(String tag) {
    final any = forms.values.map(
      (states) {
        if (forms.containsKey(tag)) {
          return !forms[tag]!.any((e) => !e.validate());
        }
        for (var state in states) {
          if (state._scope.fields.containsKey(tag)) {
            return state._scope.fields[tag]!.currentState!.validate();
          }
        }
      },
    );
    return any.contains(true);
  }

  ///Validates one, two or all fields. Returns [FormMap] success or throw.
  Map<String, dynamic> submit([List<String>? tags]) {
    final isValid = tags == null ? validate() : validateSelected(tags);
    if (!isValid) throw FormxException();
    if (form == null) log('No Formx found in the widget tree');
    return form ?? {};
  }

  ///Validates one, two or all fields. Returns [FormMap] success or null.
  Map<String, dynamic>? trySubmit([List<String>? tags]) {
    final isValid = tags == null ? validate() : validateSelected(tags);
    return isValid ? form : null;
  }

  ///Validates one. Returns [String] success or throw.
  String submitTag(String tag) {
    final isValid = validateTag(tag);
    if (!isValid) throw FormxException(tag);
    if (form == null) log('No Formx found in the widget tree');
    return isValid ? form![tag] : null;
  }

  ///Validates one. Returns [String] success or null.
  String? trySubmitTag(String tag) {
    final isValid = validateTag(tag);
    return isValid ? form![tag] : null;
  }
}

class Formx extends StatefulWidget {
  const Formx({
    super.key,
    required this.child,
    this.controller,
    this.tag,
    this.onField,
    this.fieldPadding = const EdgeInsets.all(8),
    this.onChange,
    this.onSubmit,
    this.onErrorText,
    this.decoration,
  });
  final Widget child;
  final FormxController? controller;

  ///A tag to indentify nested [Formx]. The main [Formx] tag must be null.
  ///
  ///Ex: { \
  ///     'name': '', \
  ///     'address': {...}, <- `FormX(tag: 'address')` \
  ///     'telephone': '', \
  ///    }
  final String? tag;

  ///Wrapper for each field.
  final Widget Function(String tag, Widget child)? onField;

  ///Listens to all descendants changes.
  final ValueChanged<Map<String, dynamic>>? onChange;

  ///Submits if all descendants are valid.
  final FutureOr Function(Map<String, dynamic> form)? onSubmit;

  ///Returns the field validator response;
  final String Function(String tag, String error)? onErrorText;

  ///Decorates each [Fieldx] below. You can use tag to differentiate.
  final InputDecoration Function(String tag)? decoration;

  ///The padding for each field.
  final EdgeInsets fieldPadding;

  ///Access to [FormxController].
  static FormxController of(BuildContext context) =>
      _InheritedForm.of(context).controller;

  ///Whether any descendant is loading.
  static final ValueListenable<bool> loading = _loading;
  static final _loading = ValueNotifier(false);

  @override
  State<Formx> createState() => _FormxState();
}

class _FormxState extends State<Formx> {
  final key = GlobalKey<FormState>();

  ///The current [Formx] scope.
  _InheritedForm? self;

  @override
  Widget build(BuildContext context) {
    //Tries to inherits from parent Formx.
    final parent = _InheritedForm.maybeOf(context);

    return _InheritedForm(
      form: self?.form ?? {},
      fields: self?.fields ?? {},
      controller: widget.controller ?? parent?.controller ?? FormxController(),
      onChange: (form) {
        if (widget.tag != null) parent?.form[widget.tag!] = form;
        widget.onChange?.call(form);
        parent?.onChange?.call(parent.form);
      },
      onSubmit: widget.onSubmit ?? parent?.onSubmit,
      isLoading: ValueNotifier(false),
      onErrorText: widget.onErrorText ?? parent?.onErrorText,
      decoration: widget.decoration ?? parent?.decoration,
      onField: widget.onField ?? parent?.onField,
      fieldPadding: widget.fieldPadding,
      child: Form(
        key: key,
        child: Builder(
          builder: (context) {
            self = _InheritedForm.of(context)
              ..controller
                  .states
                  .putIfAbsent(widget.tag, () => {})
                  .add(key.currentState!);
            return widget.child;
          },
        ),
      ),
    );
  }
}

// in_form

class _InheritedForm extends InheritedWidget {
  const _InheritedForm({
    required super.child,
    required this.form,
    required this.controller,
    required this.fields,
    required this.decoration,
    required this.onChange,
    required this.onSubmit,
    required this.onErrorText,
    required this.onField,
    required this.fieldPadding,
    required this.isLoading,
  });

  final ValueNotifier<bool> isLoading;
  final Map<String, dynamic> form;
  final FormxController controller;
  final FieldMap fields;
  final InputDecoration Function(String tag)? decoration;
  final ValueChanged<Map<String, dynamic>>? onChange;
  final FutureOr Function(Map<String, dynamic> form)? onSubmit;
  final String Function(String tag, String error)? onErrorText;
  final Widget Function(String tag, Widget child)? onField;
  final EdgeInsets fieldPadding;

  static _InheritedForm of(BuildContext context) {
    final maybe = maybeOf(context);
    assert(maybe != null, 'No [Formx] was found above this context.');
    return maybe!;
  }

  static _InheritedForm? maybeOf(BuildContext context) =>
      context.getInheritedWidgetOfExactType<_InheritedForm>();

  @override
  bool updateShouldNotify(_InheritedForm oldWidget) => false;
}

///A TextField that saves all the value changes in Form.of(context).form.
class Fieldx extends StatefulWidget {
  ///Creates a TextField where the value is stored in [tag].
  const Fieldx(
    this.tag, {
    this.controller,
    this.initialValue,
    this.readOnly = false,
    this.obscure = false,
    this.decoration,
    this.validator,
    this.keyboardType,
    this.mask,
    this.keepMask = false,
    this.onSubmit,
    super.key,
  });

  ///Aditionally validates if the value is empty.
  ///Returns 'form.empty.$tag' if requiredText is null.
  ///
  ///You can use [validator] for aditional validations.
  factory Fieldx.required(
    String tag, {
    String? requiredText,
    FieldController? controller,
    String? initialValue,
    bool readOnly = false,
    bool obscure = false,
    InputDecoration? decoration,
    FormFieldValidator<String>? validator,
    TextInputType? keyboardType,
    String? mask,
    bool keepMask = false,
    ValueSetter<String>? onSubmit,
    Key? key,
  }) =>
      Fieldx(
        tag,
        readOnly: readOnly,
        obscure: obscure,
        controller: controller,
        decoration: decoration,
        validator: (value) {
          if (value == null) return null;
          if (value.isEmpty) return requiredText ?? 'form.required.$tag';
          return validator?.call(value);
        },
        keyboardType: keyboardType,
        mask: mask,
        keepMask: keepMask,
        onSubmit: onSubmit,
        key: key,
      );

  ///Where to store the value in the map. Will always lowercase.
  final String tag;
  final bool readOnly;
  final bool obscure;
  final String? initialValue;
  final FutureOr Function(String)? onSubmit;
  final FieldController? controller;
  final InputDecoration? decoration;
  final FormFieldValidator<String>? validator;
  final TextInputType? keyboardType;

  ///Use # to mask numbers and A to mask letters.
  ///Ex: cpf: ###.###.###-##.
  final String? mask;

  ///If false, shows the mask only on the UI. Keeps value untouched.
  final bool keepMask;

  @override
  State<Fieldx> createState() => _FieldxState();
}

///Controls a specific [Fieldx]. Must be unique.
class FieldController {
  FieldController({String? text}) {
    this.text = text ?? '';
  }

  final key = GlobalKey<FormFieldState>();
  final focus = FocusNode();
  final edit = TextEditingController();

  ///The current field text.
  String get text => edit.text;

  ///Modifies the current field text.
  set text(String text) => edit.text = text;

  ///Request focus to the attached field.
  void requestFocus() => focus.requestFocus();

  ///Resets the attached field.
  void reset() => key.currentState?.reset();

  ///Validates this field.
  bool validate() => key.currentState?.validate() ?? false;
}

class _FieldxState extends State<Fieldx> {
  var autoValidate = AutovalidateMode.disabled;
  late var obscure = widget.obscure;
  late final controller =
      widget.controller ?? FieldController(text: scope?.form[widget.tag]);

  @override
  void dispose() {
    Formx._loading.value = false;
    super.dispose();
  }

  late final scope = _InheritedForm.maybeOf(context);

  @override
  Widget build(BuildContext context) {
    final dec = scope?.decoration?.call(widget.tag) ?? const InputDecoration();
    final decoration = dec.update(widget.decoration);

    void onChanged(String value) {
      var text = value;

      final chars = widget.mask?.removeChars('A#') ?? '';
      if (!widget.keepMask) text = value.removeChars(chars);

      //Save current value on form.
      scope?.form[widget.tag] = text;

      //Cleans map, if empty value.
      if (value.isEmpty) scope?.form.remove(widget.tag);

      //Notifies the owner.
      scope?.onChange?.call(scope!.form);
    }

    Widget obscureIcon() {
      return IconButton(
        splashRadius: 20,
        onPressed: () => setState(() => obscure = !obscure),
        icon: Icon(obscure ? CupertinoIcons.eye_slash : CupertinoIcons.eye),
      );
    }

    Widget? icon() {
      if (widget.obscure) return obscureIcon();
      return null;
    }

    final field = TextFormField(
      initialValue: widget.initialValue,
      readOnly: widget.readOnly,
      focusNode: controller.focus,
      controller: controller.edit,
      autovalidateMode: autoValidate,
      inputFormatters: [MaskTextInputFormatter(mask: widget.mask)],
      keyboardType: widget.keyboardType,
      key: scope?.fields[widget.tag] = controller.key,
      obscureText: obscure,
      obscuringCharacter: '*',
      onChanged: onChanged,
      validator: (value) {
        var text = value;

        final chars = widget.mask?.removeChars('A#') ?? '';
        if (!widget.keepMask) text = text?.removeChars(chars);

        //Callbacks to onChanged(value)
        final errorText = widget.validator?.call(text);
        if (errorText == null) return null;

        //If there is an error, starts autovalidating.
        if (autoValidate == AutovalidateMode.disabled) {
          setState(() => autoValidate = AutovalidateMode.always);
        }

        //Called only when there is an errorText.
        return scope?.onErrorText?.call(widget.tag, errorText) ?? errorText;
      },
      onFieldSubmitted: (value) async {
        try {
          scope?.isLoading.value = Formx._loading.value = true;

          if (widget.onSubmit != null && controller.validate()) {
            return await widget.onSubmit?.call(value);
          }

          if (scope?.controller.validate() ?? false) {
            await scope?.onSubmit?.call(scope!.form);
          }
        } catch (_) {
        } finally {
          scope?.isLoading.value = Formx._loading.value = false;
        }
      },
      decoration: decoration.copyWith(suffixIcon: icon()),
    );

    return Padding(
      padding: scope?.fieldPadding ?? EdgeInsets.zero,
      child: scope?.onField?.call(widget.tag, field) ?? field,
    );
  }
}

extension on FormState {
  _InheritedForm get _scope => _InheritedForm.of(context);
}

extension InputExt on InputDecoration {
  ///Updates all the fields, if any.
  InputDecoration update(InputDecoration? decoration) {
    return InputDecoration(
      icon: decoration?.icon ?? icon,
      iconColor: decoration?.iconColor ?? iconColor,
      label: decoration?.label ?? label,
      labelText: decoration?.labelText ?? labelText,
      labelStyle: decoration?.labelStyle ?? labelStyle,
      floatingLabelStyle: decoration?.floatingLabelStyle ?? floatingLabelStyle,
      helperText: decoration?.helperText ?? helperText,
      helperStyle: decoration?.helperStyle ?? helperStyle,
      helperMaxLines: decoration?.helperMaxLines ?? helperMaxLines,
      hintText: decoration?.hintText ?? hintText,
      hintStyle: decoration?.hintStyle ?? hintStyle,
      hintTextDirection: decoration?.hintTextDirection ?? hintTextDirection,
      hintMaxLines: decoration?.hintMaxLines ?? hintMaxLines,
      errorText: decoration?.errorText ?? errorText,
      errorStyle: decoration?.errorStyle ?? errorStyle,
      errorMaxLines: decoration?.errorMaxLines ?? errorMaxLines,
      floatingLabelBehavior:
          decoration?.floatingLabelBehavior ?? floatingLabelBehavior,
      floatingLabelAlignment:
          decoration?.floatingLabelAlignment ?? floatingLabelAlignment,
      isCollapsed: decoration?.isCollapsed ?? isCollapsed,
      isDense: decoration?.isDense ?? isDense,
      contentPadding: decoration?.contentPadding ?? contentPadding,
      prefixIcon: decoration?.prefixIcon ?? prefixIcon,
      prefix: decoration?.prefix ?? prefix,
      prefixText: decoration?.prefixText ?? prefixText,
      prefixStyle: decoration?.prefixStyle ?? prefixStyle,
      prefixIconColor: decoration?.prefixIconColor ?? prefixIconColor,
      prefixIconConstraints:
          decoration?.prefixIconConstraints ?? prefixIconConstraints,
      suffixIcon: decoration?.suffixIcon ?? suffixIcon,
      suffix: decoration?.suffix ?? suffix,
      suffixText: decoration?.suffixText ?? suffixText,
      suffixStyle: decoration?.suffixStyle ?? suffixStyle,
      suffixIconColor: decoration?.suffixIconColor ?? suffixIconColor,
      suffixIconConstraints:
          decoration?.suffixIconConstraints ?? suffixIconConstraints,
      counter: decoration?.counter ?? counter,
      counterText: decoration?.counterText ?? counterText,
      counterStyle: decoration?.counterStyle ?? counterStyle,
      filled: decoration?.filled ?? filled,
      fillColor: decoration?.fillColor ?? fillColor,
      focusColor: decoration?.focusColor ?? focusColor,
      hoverColor: decoration?.hoverColor ?? hoverColor,
      errorBorder: decoration?.errorBorder ?? errorBorder,
      focusedBorder: decoration?.focusedBorder ?? focusedBorder,
      focusedErrorBorder: decoration?.focusedErrorBorder ?? focusedErrorBorder,
      disabledBorder: decoration?.disabledBorder ?? disabledBorder,
      enabledBorder: decoration?.enabledBorder ?? enabledBorder,
      border: decoration?.border ?? border,
      enabled: decoration?.enabled ?? enabled,
      semanticCounterText:
          decoration?.semanticCounterText ?? semanticCounterText,
      alignLabelWithHint: decoration?.alignLabelWithHint ?? alignLabelWithHint,
      constraints: decoration?.constraints ?? constraints,
    );
  }
}

enum MaskAutoCompletionType {
  lazy,
  eager,
}

class MaskTextInputFormatter implements TextInputFormatter {
  final MaskAutoCompletionType type;

  String? _mask;
  List<String> _maskChars = [];
  Map<String, RegExp>? _maskFilter;

  int _maskLength = 0;
  final _TextMatcher _resultTextArray = _TextMatcher();
  String _resultTextMasked = '';

  TextEditingValue? _lastResValue;
  TextEditingValue? _lastNewValue;

  /// Create the [mask] formatter for TextField
  ///
  /// The keys of the [filter] assign which character in the mask should be
  /// replaced and the values validate the entered character
  /// By default `#` match to the number and `A` to the letter
  ///
  /// Set [type] for autocompletion behavior:
  ///  - [MaskAutoCompletionType.lazy] (default): autocomplete unfiltered
  /// characters once the following filtered character is input.
  ///  For example, with the mask "#/#" and the sequence of characters "1" then
  /// "2", the formatter will output "1", then "1/2"
  ///  - [MaskAutoCompletionType.eager]: autocomplete unfiltered characters when
  ///  the previous filtered character is input.
  ///  For example, with the mask "#/#" and the sequence of characters "1" then
  /// "2", the formatter will output "1/", then "1/2"
  MaskTextInputFormatter({
    String? mask,
    Map<String, RegExp>? filter,
    String? initialText,
    this.type = MaskAutoCompletionType.lazy,
  }) {
    updateMask(
      mask: mask,
      filter: filter ?? {'#': RegExp('[0-9]'), 'A': RegExp('[^0-9]')},
    );
    if (initialText != null) {
      formatEditUpdate(
        TextEditingValue.empty,
        TextEditingValue(text: initialText),
      );
    }
  }

  /// Change the mask
  TextEditingValue updateMask({String? mask, Map<String, RegExp>? filter}) {
    _mask = mask;
    if (filter != null) {
      _updateFilter(filter);
    }
    _calcMaskLength();
    final unmaskedText = getUnmaskedText();
    clear();
    return formatEditUpdate(
      TextEditingValue.empty,
      TextEditingValue(
        text: unmaskedText,
        selection: TextSelection.collapsed(offset: unmaskedText.length),
      ),
    );
  }

  /// Get current mask
  String? getMask() {
    return _mask;
  }

  /// Get masked text, e.g. "+0 (123) 456-78-90"
  String getMaskedText() {
    return _resultTextMasked;
  }

  /// Get unmasked text, e.g. "01234567890"
  String getUnmaskedText() {
    return _resultTextArray.toString();
  }

  /// Check if target mask is filled
  bool isFill() {
    return _resultTextArray.length == _maskLength;
  }

  /// Clear masked text of the formatter
  /// Note: you need to call this method if you clear the text of the TextField
  /// because it doesn't call the formatter when it has empty text
  void clear() {
    _resultTextMasked = '';
    _resultTextArray.clear();
    _lastResValue = null;
    _lastNewValue = null;
  }

  /// Mask some text
  String maskText(String text) {
    return MaskTextInputFormatter(
      mask: _mask,
      filter: _maskFilter,
      initialText: text,
    ).getMaskedText();
  }

  /// Unmask some text
  String unmaskText(String text) {
    return MaskTextInputFormatter(
      mask: _mask,
      filter: _maskFilter,
      initialText: text,
    ).getUnmaskedText();
  }

  @override
  TextEditingValue formatEditUpdate(
    TextEditingValue oldValue,
    TextEditingValue newValue,
  ) {
    if (_lastResValue == oldValue && newValue == _lastNewValue) {
      return oldValue;
    }
    if (oldValue.text.isEmpty) {
      _resultTextArray.clear();
    }
    _lastNewValue = newValue;
    return _lastResValue = _format(oldValue, newValue);
  }

  TextEditingValue _format(
    TextEditingValue oldValue,
    TextEditingValue newValue,
  ) {
    final mask = _mask;

    if (mask == null || mask.isEmpty == true) {
      _resultTextMasked = newValue.text;
      _resultTextArray.set(newValue.text);
      return newValue;
    }

    final beforeText = oldValue.text;
    final afterText = newValue.text;

    final beforeSelection = oldValue.selection;
    final afterSelection = newValue.selection;

    final beforeSelectionStart = afterSelection.isValid
        ? beforeSelection.isValid
            ? beforeSelection.start
            : 0
        : 0;
    final beforeSelectionLength = afterSelection.isValid
        ? beforeSelection.isValid
            ? beforeSelection.end - beforeSelection.start
            : 0
        : oldValue.text.length;

    final lengthDifference =
        afterText.length - (beforeText.length - beforeSelectionLength);
    final lengthRemoved = lengthDifference < 0 ? lengthDifference.abs() : 0;
    final lengthAdded = lengthDifference > 0 ? lengthDifference : 0;

    final afterChangeStart = math.max(0, beforeSelectionStart - lengthRemoved);
    final afterChangeEnd = math.max(0, afterChangeStart + lengthAdded);

    final beforeReplaceStart =
        math.max(0, beforeSelectionStart - lengthRemoved);
    final beforeReplaceLength = beforeSelectionLength + lengthRemoved;

    final beforeResultTextLength = _resultTextArray.length;

    var currentResultTextLength = _resultTextArray.length;
    var currentResultSelectionStart = 0;
    var currentResultSelectionLength = 0;

    for (var i = 0;
        i < math.min(beforeReplaceStart + beforeReplaceLength, mask.length);
        i++) {
      if (_maskChars.contains(mask[i]) && currentResultTextLength > 0) {
        currentResultTextLength -= 1;
        if (i < beforeReplaceStart) {
          currentResultSelectionStart += 1;
        }
        if (i >= beforeReplaceStart) {
          currentResultSelectionLength += 1;
        }
      }
    }

    final replacementText =
        afterText.substring(afterChangeStart, afterChangeEnd);
    var targetCursorPosition = currentResultSelectionStart;
    if (replacementText.isEmpty) {
      _resultTextArray.removeRange(
        currentResultSelectionStart,
        currentResultSelectionStart + currentResultSelectionLength,
      );
    } else {
      if (currentResultSelectionLength > 0) {
        _resultTextArray.removeRange(
          currentResultSelectionStart,
          currentResultSelectionStart + currentResultSelectionLength,
        );
      }
      _resultTextArray.insert(currentResultSelectionStart, replacementText);
      targetCursorPosition += replacementText.length;
    }

    if (beforeResultTextLength == 0 && _resultTextArray.length > 1) {
      for (var i = 0; i < mask.length; i++) {
        if (_maskChars.contains(mask[i])) {
          final resultPrefix = _resultTextArray._symbolArray.take(i).toList();
          for (var j = 0; j < resultPrefix.length; j++) {
            if (_resultTextArray.length <= j ||
                (mask[j] != resultPrefix[j] ||
                    (mask[j] == resultPrefix[j] &&
                        j == resultPrefix.length - 1))) {
              _resultTextArray.removeRange(0, j);
              break;
            }
          }
          break;
        }
      }
    }

    var curTextPos = 0;
    var maskPos = 0;
    _resultTextMasked = '';
    var cursorPos = -1;
    var nonMaskedCount = 0;

    while (maskPos < mask.length) {
      final curMaskChar = mask[maskPos];
      final isMaskChar = _maskChars.contains(curMaskChar);

      var curTextInRange = curTextPos < _resultTextArray.length;

      String? curTextChar;
      if (isMaskChar && curTextInRange) {
        while (curTextChar == null && curTextInRange) {
          final potentialTextChar = _resultTextArray[curTextPos];
          if (_maskFilter?[curMaskChar]?.hasMatch(potentialTextChar) ?? false) {
            curTextChar = potentialTextChar;
          } else {
            _resultTextArray.removeAt(curTextPos);
            curTextInRange = curTextPos < _resultTextArray.length;
            if (curTextPos <= targetCursorPosition) {
              targetCursorPosition -= 1;
            }
          }
        }
      } else if (!isMaskChar &&
          !curTextInRange &&
          type == MaskAutoCompletionType.eager) {
        curTextInRange = true;
      }

      if (isMaskChar && curTextInRange && curTextChar != null) {
        _resultTextMasked += curTextChar;
        if (curTextPos == targetCursorPosition && cursorPos == -1) {
          cursorPos = maskPos - nonMaskedCount;
        }
        nonMaskedCount = 0;
        curTextPos += 1;
      } else {
        if (curTextPos == targetCursorPosition &&
            cursorPos == -1 &&
            !curTextInRange) {
          cursorPos = maskPos;
        }

        if (!curTextInRange) {
          break;
        } else {
          _resultTextMasked += mask[maskPos];
        }

        if (type == MaskAutoCompletionType.lazy || lengthRemoved > 0) {
          nonMaskedCount++;
        }
      }

      maskPos += 1;
    }

    if (nonMaskedCount > 0) {
      _resultTextMasked = _resultTextMasked.substring(
        0,
        _resultTextMasked.length - nonMaskedCount,
      );
      cursorPos -= nonMaskedCount;
    }

    if (_resultTextArray.length > _maskLength) {
      _resultTextArray.removeRange(_maskLength, _resultTextArray.length);
    }

    final finalCursorPosition =
        cursorPos < 0 ? _resultTextMasked.length : cursorPos;

    return TextEditingValue(
      text: _resultTextMasked,
      selection: TextSelection(
        baseOffset: finalCursorPosition,
        extentOffset: finalCursorPosition,
        affinity: newValue.selection.affinity,
        isDirectional: newValue.selection.isDirectional,
      ),
    );
  }

  void _calcMaskLength() {
    _maskLength = 0;
    final mask = _mask;
    if (mask != null) {
      for (var i = 0; i < mask.length; i++) {
        if (_maskChars.contains(mask[i])) {
          _maskLength++;
        }
      }
    }
  }

  void _updateFilter(Map<String, RegExp> filter) {
    _maskFilter = filter;
    _maskChars = _maskFilter?.keys.toList(growable: false) ?? [];
  }
}

class _TextMatcher {
  final List<String> _symbolArray = <String>[];
  int get length => _symbolArray.fold(0, (prev, match) => prev + match.length);
  void removeRange(int start, int end) => _symbolArray.removeRange(start, end);
  void insert(int start, String substring) {
    for (var i = 0; i < substring.length; i++) {
      _symbolArray.insert(start + i, substring[i]);
    }
  }

  bool get isEmpty => _symbolArray.isEmpty;
  void removeAt(int index) => _symbolArray.removeAt(index);
  String operator [](int index) => _symbolArray[index];
  void clear() => _symbolArray.clear();
  @override
  String toString() => _symbolArray.join();

  void set(String text) {
    _symbolArray.clear();
    for (var i = 0; i < text.length; i++) {
      _symbolArray.add(text[i]);
    }
  }
}

class AppDropdownInput<T> extends StatelessWidget {
  final String hintText;
  final List<T> options;
  final T value;
  final String Function(T)? getLabel;
  final void Function(T?)? onChanged;

  const AppDropdownInput({
    this.hintText = 'Please select an Option',
    this.options = const [],
    this.getLabel,
    required this.value,
    this.onChanged,
    super.key,
  });

  @override
  Widget build(BuildContext context) {
    return FormField<T>(
      builder: (state) {
        return InputDecorator(
          decoration: InputDecoration(
            contentPadding:
                const EdgeInsets.symmetric(horizontal: 20.0, vertical: 15.0),
            labelText: hintText,
            border:
                OutlineInputBorder(borderRadius: BorderRadius.circular(5.0)),
          ),
          isEmpty: value == null || value == '',
          child: DropdownButtonHideUnderline(
            child: DropdownButton<T>(
              value: value,
              isDense: true,
              onChanged: onChanged,
              items: options.map((value) {
                return DropdownMenuItem<T>(
                  value: value,
                  child: Text(getLabel?.call(value) ?? value.toString()),
                );
              }).toList(),
            ),
          ),
        );
      },
    );
  }
}

extension StrinExt on String {
  /// Capitalizes all words. ex: your name => Your Name.
  String get capitalized => split(' ').map((e) => e.capitalizeFirst).join(' ');

  /// Uppercase first letter inside string and let the others lowercase.
  /// Example: your name => Your name.
  String get capitalizeFirst {
    if (isEmpty) return '';
    return this[0].toUpperCase() + substring(1).toLowerCase();
  }

  String get lastPascalCaseWord {
    final pattern = RegExp(r'[A-Z][a-z0-9]*(?=\b)');
    final matches = pattern.allMatches(this);
    if (matches.isNotEmpty) {
      final match = matches.last;
      return substring(match.start, match.end);
    }
    return this;
  }

  ///All sub words between [pattern].
  ///Ex: 'a.b.c' -> ['a.b.c','a.b','a'].
  List<String> subWords(Pattern pattern) {
    final words = split(pattern);
    final nestedStrings = <String>[];

    for (var i = 0; i < words.length; i++) {
      final currentNestedString = words.sublist(0, i + 1).join('.');
      nestedStrings.add(currentNestedString);
    }

    return nestedStrings.reversed.toList();
  } //tested [string_test]

  ///Remove all the chars in [charsToRemove].
  String removeChars(String charsToRemove) {
    var result = this;
    charsToRemove.split('').forEach((char) {
      result = result.replaceAll(char, '');
    });
    return result;
  }
}

extension MapMergeExtension on Map {
  /// Splits [path] into a nested map.
  Map<String, dynamic> nestIn(String path, {Pattern pattern = '.'}) {
    final keys = path.split(pattern);
    var nestedMap = <String, dynamic>{};
    var currentMap = nestedMap;

    for (var i = 0; i < keys.length; i++) {
      if (i < keys.length - 1) {
        currentMap[keys[i]] = <String, dynamic>{};
        currentMap = currentMap[keys[i]] as Map<String, dynamic>;
      } else {
        currentMap[keys[i]] = this;
      }
    }

    return nestedMap;
  }

  /// Merges [other] into this map, recursively.
  ///
  /// Same as [Map.addAll] but never overwrites.
  void merge(Map other) {
    other.forEach((key, value) {
      if (value is Map && this[key] is Map) {
        (this[key] as Map).merge(value);
      } else {
        this[key] = value;
      }
    });
  }
}

extension Validatorx on String {
  bool get hasUppercase => RegExp('[A-Z]').hasMatch(this);
  bool get hasLowercase => RegExp('[a-z]').hasMatch(this);
  bool get hasNumber => RegExp('[0-9]').hasMatch(this);
  bool get hasSpecial => RegExp(r'[!@#\$&*~]').hasMatch(this);
  bool get isLength8 => length >= 6;
  bool get isPassword => hasNumber && isLength8;
  bool get isEmail => Validationx.isEmail(this);
  bool get isUrl => Validationx.isUrl(this);
  bool get isCpf => Validationx.isCpf(this);
  bool get isCnpj => Validationx.isCnpj(this);
  bool get isPhone => Validationx.isPhone(this);
  bool get isCurrency => Validationx.isCurrency(this);
  bool get isDateTime => Validationx.isDateTime(this);
  bool get isPassport => Validationx.isPassport(this);
  bool get isBinary => Validationx.isBinary(this);
  bool get isHexadecimal => Validationx.isHexadecimal(this);

  static String? password(String? text) => text?.test(text.isPassword, true);
  static String? email(String? text) => text?.test(text.isEmail, true);
  static String? cpf(String? text) => text?.test(text.isCpf, true);

  ///Validates if the field is required. Nothing else.
  static String? required(String? value) => value?.validator(required: true);

  ///Short version.
  String? test(bool test, [bool required = false]) {
    return validator(test: test, required: required);
  }

  ///Declares a Form validator callback.
  String? validator({
    bool? test,
    bool required = false,
    String invalidText = 'form.invalid',
    String requiredText = 'form.required',
  }) {
    if (isEmpty && required) return requiredText;
    if (test == null) return null;
    return test == true ? null : invalidText;
  }
}

@Deprecated('Use Validations instead')
typedef Utils = Validationx;

mixin Validationx {
  static bool isEmail(String value) => RegExp(
        r'^(([^<>()[\]\\.,;:\s@\"]+(\.[^<>()[\]\\.,;:\s@\"]+)*)|(\".+\"))@((\[[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\])|(([a-zA-Z\-0-9]+\.)+[a-zA-Z]{2,}))$',
      ).hasMatch(value);

  static bool isUrl(String value) => RegExp(
        r"^((((H|h)(T|t)|(F|f))(T|t)(P|p)((S|s)?))\://)?(www.|[a-zA-Z0-9].)[a-zA-Z0-9\-\.]+\.[a-zA-Z]{2,6}(\:[0-9]{1,5})*(/($|[a-zA-Z0-9\.\,\;\?\'\\\+&amp;%\$#\=~_\-]+))*$",
      ).hasMatch(value);

  static bool isPhone(String value) {
    if (value.length > 16 || value.length < 9) return false;
    return RegExp(
      r'^[+]*[(]{0,1}[0-9]{1,4}[)]{0,1}[-\s\./0-9]*$',
    ).hasMatch(value);
  }

  static bool isDateTime(String value) =>
      RegExp(r'^\d{4}-\d{2}-\d{2}[ T]\d{2}:\d{2}:\d{2}.\d{3}Z?$')
          .hasMatch(value);

  static bool isPassport(String value) =>
      RegExp(r'^(?!^0+$)[a-zA-Z0-9]{6,9}$').hasMatch(value);

  static bool isBinary(String value) => RegExp(r'^[0-1]+$').hasMatch(value);

  static bool isHexadecimal(String value) =>
      RegExp(r'^#?([0-9a-fA-F]{3}|[0-9a-fA-F]{6})$').hasMatch(value);

  static bool isCurrency(String value) => RegExp(
        r'^(S?\$|\₩|Rp|\¥|\€|\₹|\₽|fr|R\$|R)?[ ]?[-]?([0-9]{1,3}[,.]([0-9]{3}[,.])*[0-9]{3}|[0-9]+)([,.][0-9]{1,2})?( ?(USD?|AUD|NZD|CAD|CHF|GBP|CNY|EUR|JPY|IDR|MXN|NOK|KRW|TRY|INR|RUB|BRL|ZAR|SGD|MYR))?$',
      ).hasMatch(value);

  //Check if num is a cnpj
  static bool isCnpj(String cnpj) {
    // Obter somente os números do CNPJ
    final numbers = cnpj.replaceAll(RegExp('[^0-9]'), '');

    // Testar se o CNPJ possui 14 dígitos
    if (numbers.length != 14) return false;

    // Testar se todos os dígitos do CNPJ são iguais
    if (RegExp(r'^(\d)\1*$').hasMatch(numbers)) {
      return false;
    }

    // Dividir dígitos
    final digits = numbers.split('').map(int.parse).toList();

    // Calcular o primeiro dígito verificador
    var calcDv1 = 0;
    var j = 0;
    for (final i in Iterable<int>.generate(12, (i) => i < 4 ? 5 - i : 13 - i)) {
      calcDv1 += digits[j++] * i;
    }
    calcDv1 %= 11;
    final dv1 = calcDv1 < 2 ? 0 : 11 - calcDv1;

    // Testar o primeiro dígito verificado
    if (digits[12] != dv1) {
      return false;
    }

    // Calcular o segundo dígito verificador
    var calcDv2 = 0;
    j = 0;
    for (final i in Iterable<int>.generate(13, (i) => i < 5 ? 6 - i : 14 - i)) {
      calcDv2 += digits[j++] * i;
    }
    calcDv2 %= 11;
    final dv2 = calcDv2 < 2 ? 0 : 11 - calcDv2;

    // Testar o segundo dígito verificador
    if (digits[13] != dv2) {
      return false;
    }

    return true;
  }

  /// Checks if the cpf is valid.
  static bool isCpf(String cpf) {
    // get only the numbers
    final numbers = cpf.replaceAll(RegExp('[^0-9]'), '');
    // Test if the CPF has 11 digits
    if (numbers.length != 11) {
      return false;
    }
    // Test if all CPF digits are the same
    if (RegExp(r'^(\d)\1*$').hasMatch(numbers)) {
      return false;
    }

    // split the digits
    final digits = numbers.split('').map(int.parse).toList();

    // Calculate the first verifier digit
    var calcDv1 = 0;
    for (final i in Iterable<int>.generate(9, (i) => 10 - i)) {
      calcDv1 += digits[10 - i] * i;
    }
    calcDv1 %= 11;

    final dv1 = calcDv1 < 2 ? 0 : 11 - calcDv1;

    // Tests the first verifier digit
    if (digits[9] != dv1) {
      return false;
    }

    // Calculate the second verifier digit
    var calcDv2 = 0;
    for (final i in Iterable<int>.generate(10, (i) => 11 - i)) {
      calcDv2 += digits[11 - i] * i;
    }
    calcDv2 %= 11;

    final dv2 = calcDv2 < 2 ? 0 : 11 - calcDv2;

    // Test the second verifier digit
    if (digits[10] != dv2) {
      return false;
    }
    return true;
  }
}
